from __future__ import annotations

from collections.abc import MutableMapping
from datetime import datetime
from enum import Enum

import jsonref
from pydantic import BaseModel, model_serializer

from solaredge2mqtt import __version__


class EnumModel(Enum):
    def __new__(cls, *args):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __str__(self) -> str:
        return str(self._value_)

    def __repr__(self) -> str:
        return str(self._value_)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.value == other.value

    def __hash__(self):
        return hash(self.value)

    @classmethod
    def from_string(cls, value: str) -> EnumModel:
        for item in cls.__members__.values():
            if item.value == value:
                return item
        raise ValueError(f"No enum value {value} found.")

    @model_serializer
    def serialize(self) -> any:
        return self.value


class Solaredge2MQTTBaseModel(BaseModel):
    def model_dump_influxdb(self, exclude: list[str] | None = None) -> dict[str, any]:
        return self._flatten_dict(self.model_dump(exclude=exclude, exclude_none=True))

    def _flatten_dict(
        self, d: MutableMapping, join_chr: str = "_", parent_key: str = ""
    ) -> MutableMapping:
        items = []
        for k, v in d.items():
            new_key = parent_key + join_chr + k if parent_key else k
            if isinstance(v, MutableMapping):
                items.extend(self._flatten_dict(v, join_chr, new_key).items())
            else:
                if isinstance(v, int):
                    v = float(v)
                elif isinstance(v, datetime):
                    v = v.isoformat()

                items.append((new_key, v))
        return dict(items)

    @classmethod
    def _walk_schema_for_homeassistant_entities(
        cls,
        properties: dict[str, dict],
        parent_jsonpath: str | None = None,
        parent_name: str | None = None,
    ) -> list[dict]:
        items: list[dict] = []
        for key, prop in properties.items():
            new_jsonpath = parent_jsonpath + "." + key if parent_jsonpath else key
            new_name = (
                parent_name + " " + prop["title"] if parent_name else prop["title"]
            )
            if "properties" in prop:
                items.extend(
                    cls._walk_schema_for_homeassistant_entities(
                        prop["properties"], new_jsonpath, new_name
                    )
                )
            elif "allOf" in prop:
                items.extend(
                    cls._walk_schema_for_homeassistant_entities(
                        prop["allOf"][0]["properties"], new_jsonpath, new_name
                    )
                )
            elif "anyOf" in prop and "properties" in prop["anyOf"][0]:
                items.extend(
                    cls._walk_schema_for_homeassistant_entities(
                        prop["anyOf"][0]["properties"], new_jsonpath, new_name
                    )
                )
            else:
                entity: dict = {"name": new_name, "jsonpath": new_jsonpath}

                if "ha_type" in prop:
                    entity["ha_type"] = prop["ha_type"]
                    entity["icon"] = prop["icon"]
                    items.append(entity)

        return items

    def _default_homeassistant_device_info(self, name: str) -> dict[str, any]:
        return {
            "name": f"SolarEdge2MQTT {name}",
            "manufacturer": "DerOetzi",
            "model": "SolarEdge2MQTT",
            "sw_version": __version__,
        }

    @classmethod
    def homeassistant_entities_info(cls) -> list[dict]:
        return cls._walk_schema_for_homeassistant_entities(
            jsonref.replace_refs(cls.model_json_schema(mode="serialization"))[
                "properties"
            ]
        )






from __future__ import annotations

import time
from asyncio import Event, to_thread
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

from numpy import percentile
from pandas import DataFrame
from sklearn import clone
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.inspection import permutation_importance
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit, train_test_split
from sklearn.pipeline import Pipeline
from tzlocal import get_localzone_name

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.influxdb import InfluxDB, Point
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import Interval10MinTriggerEvent
from solaredge2mqtt.services.forecast.encoders import (
    CategoricalEncoder,
    CyclicalEncoder,
    SunEncoder,
    TimeEncoder,
)
from solaredge2mqtt.services.forecast.events import ForecastEvent
from solaredge2mqtt.services.forecast.models import Forecast, ForecasterType
from solaredge2mqtt.services.forecast.settings import ForecastSettings
from solaredge2mqtt.services.weather.events import WeatherUpdateEvent
from solaredge2mqtt.services.weather.models import OpenWeatherMapForecastData

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings.models import LocationSettings


LOCAL_TZ = get_localzone_name()


class ForecastService:
    def __init__(
        self,
        settings: ForecastSettings,
        location: LocationSettings,
        event_bus: EventBus,
        influxdb: InfluxDB,
    ) -> None:
        self.settings = settings
        self.location = location
        self.event_bus = event_bus
        self._subscribe_events()

        self.influxdb = influxdb

        self.forecasters: dict[ForecasterType, Forecaster] = {
            typed: Forecaster(typed, location, settings.hyperparametertuning)
            for typed in ForecasterType
        }

        self.last_weather_forecast: list[OpenWeatherMapForecastData] | None = None
        self.last_hour_forecast: dict[int, OpenWeatherMapForecastData] | None = None

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(WeatherUpdateEvent, self.weather_update)
        self.event_bus.subscribe(Interval10MinTriggerEvent, self.forecast_loop)

    async def weather_update(self, event: WeatherUpdateEvent) -> None:
        self.last_weather_forecast = event.weather.hourly

        if self.last_hour_forecast is None:
            self.last_hour_forecast = {}

        self.last_hour_forecast[event.weather.hourly[0].hour] = event.weather.hourly[0]

        now = datetime.now().astimezone()
        last_hour = now - timedelta(hours=1)

        for delete_hour in range(0, 24):
            if delete_hour in [now.hour, last_hour.hour]:
                continue

            self.last_hour_forecast.pop(delete_hour, None)

        logger.debug(self.last_hour_forecast)

        if last_hour.hour in self.last_hour_forecast:
            await self.write_new_training_data(self.last_hour_forecast[last_hour.hour])

    async def write_new_training_data(
        self, last_hour_weather_forecast: OpenWeatherMapForecastData
    ) -> None:
        now = datetime.now().astimezone()
        last_hour = now.replace(minute=0, second=0, microsecond=0) - timedelta(hours=1)

        training_data = last_hour_weather_forecast.model_dump_estimation_data()
        training_data["time"] = last_hour
        training_data = self.add_last_hour_pv_production(training_data)
        self.write_new_training_data_to_influxdb(training_data)

        if (now.minute // 10) * 10 == 20:
            await self.train()

    def add_last_hour_pv_production(
        self, trainings_data
    ) -> dict[str, str | float | int | None]:
        production_data = self.influxdb.query_first("production")

        if production_data is None:
            raise InvalidDataException(
                "Missing production data of last hour for forecast training."
                + " Did the service write power information to InfluxDB?"
            )

        trainings_data[ForecasterType.POWER.target_column] = round(
            production_data[ForecasterType.POWER.target_column]
        )
        trainings_data[ForecasterType.ENERGY.target_column] = round(
            production_data[ForecasterType.ENERGY.target_column], 2
        )
        return trainings_data

    def write_new_training_data_to_influxdb(self, trainings_data):
        point = Point("forecast_training")
        for key, value in trainings_data.items():
            if isinstance(value, (int, float, str, bool)):
                point.field(key, value)

        point.time(trainings_data["time"].astimezone(timezone.utc))

        logger.info("Write new forecast training data to influxdb")
        logger.debug(trainings_data)
        self.influxdb.write_point(point)

    async def train(self) -> None:
        data = await self.influxdb.query_dataframe("training_data")
        data["time"] = data["_time"].dt.tz_convert(LOCAL_TZ)
        await to_thread(self.training, data)

    def training(self, data: DataFrame) -> None:
        for forecaster in self.forecasters.values():
            forecaster.train(data)

    async def forecast_loop(self, _):
        if (
            not self.forecasters[ForecasterType.ENERGY].is_trained
            or not self.forecasters[ForecasterType.POWER].is_trained
        ):
            await self.train()

        if self.last_weather_forecast is None:
            raise InvalidDataException(
                "Missing weather forecast for production forecast"
            )

        estimation_data_list = [
            {
                "time": datetime(
                    year=weather_forecast.year,
                    month=weather_forecast.month,
                    day=weather_forecast.day,
                    hour=weather_forecast.hour,
                    minute=0,
                    second=0,
                    microsecond=0,
                ).astimezone(),
                **weather_forecast.model_dump_estimation_data(),
            }
            for weather_forecast in self.last_weather_forecast
        ]

        data = DataFrame(estimation_data_list)
        data["time"] = data["time"].astype(f"datetime64[ns, {LOCAL_TZ}]")

        for typed, forecaster in self.forecasters.items():
            predicted_data = await forecaster.predict(data)
            self._write_periods_to_influxdb(predicted_data, typed)

        await self.publish_forecast()

    def _write_periods_to_influxdb(
        self, periods: DataFrame, typed: ForecasterType
    ) -> None:
        points = []
        for _, period in periods.iterrows():
            point = Point("forecast")
            point.field(typed.target_column, period[typed.target_column])
            utc_time = period["time"].astimezone(timezone.utc)
            point.time(utc_time)
            points.append(point)

        logger.info("Write forecast data to influxdb")
        self.influxdb.write_points(points)

    async def publish_forecast(self) -> None:
        forecast_data = await self.influxdb.query_dataframe("forecast")
        if not forecast_data.empty:
            forecast_data["time"] = forecast_data["_time"].dt.tz_convert(LOCAL_TZ)
            power_hours = {
                row["time"]: row["power"] for idx, row in forecast_data.iterrows()
            }
            energy_hours = {
                row["time"]: round(row["energy"] * 1000)
                for idx, row in forecast_data.iterrows()
            }
            forecast = Forecast(power_period=power_hours, energy_period=energy_hours)
            logger.debug(forecast)

            await self.event_bus.emit(MQTTPublishEvent(forecast.mqtt_topic(), forecast))
            await self.event_bus.emit(ForecastEvent(forecast))


class Forecaster:
    NUMERIC_FEATURES: list[str] = [
        "clouds",
        "dew_point",
        "feels_like",
        "humidity",
        "pop",
        "pressure",
        "rain",
        "temp",
        "uvi",
        "visibility",
        "wind_speed",
        "wind_gust",
    ]
    CATEGORICAL_FEATURES: list[str] = ["weather_id", "weather_main"]
    CYCLICAL_FEATURES: dict[str, int] = {
        "wind_deg": 360,
    }
    TARGET_FEATURES: dict[str, str] = {
        ForecasterType.ENERGY.target_column: "energy",
        ForecasterType.POWER.target_column: "power",
    }

    def __init__(
        self,
        typed: ForecasterType,
        location: LocationSettings,
        enable_hyperparameter_tuning: bool = False,
    ) -> None:
        self.typed: ForecasterType = typed
        self.location = location
        self.enable_hyperparameter_tuning = enable_hyperparameter_tuning
        self.model_pipeline: Pipeline = None
        self.training_completed: Event = Event()

    def train(self, data: DataFrame) -> None:
        data_count = len(data)
        logger.info(
            f"Training model {self.typed} with {data_count} hours of data points"
        )

        if data_count < 60:
            raise InvalidDataException(
                "Forecast needs at least 60 hours of data at least to start training",
            )

        self.training_completed.clear()
        start_time = time.time()
        y_vector = data[self.typed.target_column]

        pipeline = self._prepare_model_pipeline(data.columns.to_list())

        if self.enable_hyperparameter_tuning:
            self.model_pipeline = self._hyperparametertuning(data, y_vector, pipeline)
        else:
            self.model_pipeline = pipeline

        self.model_pipeline.fit(data, y_vector)

        execution_time = time.time() - start_time
        self.training_completed.set()

        transformed_features = self.model_pipeline[
            "preprocessor"
        ].get_feature_names_out()
        logger.debug(
            "Transformed features ({count}): {features} ",
            count=len(transformed_features),
            features=", ".join(transformed_features),
        )

        selected_features = self.model_pipeline[
            "feature_selector"
        ].important_features_.tolist()
        logger.info(
            "Selected features ({count}): {features} ",
            count=len(selected_features),
            features=", ".join(selected_features),
        )

        logger.info(f"Training execution time: {execution_time:.2f} seconds")

    def _hyperparametertuning(self, data, y_vector, pipeline):
        param_grid = {
            "model__max_iter": [100, 200, 300],
            "model__max_depth": [None, 5, 10],
            "model__learning_rate": [0.01, 0.1],
        }

        grid_search = GridSearchCV(
            estimator=pipeline,
            param_grid=param_grid,
            cv=TimeSeriesSplit(n_splits=2),
            scoring="neg_mean_squared_error",
        )
        grid_search.fit(data, y_vector)

        logger.info(
            "Training with best parameters: {params}", params=grid_search.best_params_
        )
        logger.info("Training with best score: {score}", score=grid_search.best_score_)

        return clone(grid_search.best_estimator_)

    def _prepare_model_pipeline(self, x_vector_columns: list[str]) -> None:
        return Pipeline(
            steps=[
                ("preprocessor", self._prepare_preprocessor(x_vector_columns)),
                (
                    "feature_selector",
                    PFISelector(
                        estimator=HistGradientBoostingRegressor(
                            random_state=42, categorical_features="from_dtype"
                        )
                    ),
                ),
                (
                    "model",
                    HistGradientBoostingRegressor(
                        random_state=42, categorical_features="from_dtype"
                    ),
                ),
            ]
        )

    def _prepare_preprocessor(self, x_vector_columns: list[str]) -> ColumnTransformer:
        ct = ColumnTransformer(
            transformers=[
                (
                    "cyc",
                    CyclicalEncoder(**self.CYCLICAL_FEATURES),
                    self._extract_used_columns(
                        self.CYCLICAL_FEATURES, x_vector_columns
                    ),
                ),
                (
                    "num",
                    "passthrough",
                    self._extract_used_columns(self.NUMERIC_FEATURES, x_vector_columns),
                ),
                (
                    "time",
                    TimeEncoder(),
                    ["time"],
                ),
                ("sun", SunEncoder(self.location), ["time"]),
                (
                    "cat",
                    CategoricalEncoder(),
                    self._extract_used_columns(
                        self.CATEGORICAL_FEATURES, x_vector_columns
                    ),
                ),
            ],
            remainder="drop",
        )
        ct.set_output(transform="pandas")
        return ct

    @staticmethod
    def _extract_used_columns(
        typed_features: list[str] | dict[str, int], x_vector_columns: list[str]
    ) -> list[str]:
        return [col for col in typed_features if col in x_vector_columns]

    @property
    def is_trained(self) -> bool:
        return self.model_pipeline is not None

    async def predict(self, new_data: DataFrame) -> DataFrame:
        if self.model_pipeline is None:
            raise InvalidDataException("The model has not been trained yet.")

        data_for_prediction = new_data.copy()

        await self.training_completed.wait()

        predictions = self.model_pipeline.predict(data_for_prediction)
        data_for_prediction[self.typed.target_column] = predictions
        data_for_prediction[self.typed.target_column] = data_for_prediction[
            self.typed.target_column
        ].apply(self.typed.prepare_value)

        return data_for_prediction


class PFISelector(BaseEstimator, TransformerMixin):
    def __init__(self, estimator, n_repeats=10):
        self.estimator = estimator
        self.n_repeats = n_repeats

        self.estimator_ = None
        self.feature_importances_ = None
        self.important_indices_ = None
        self.important_features_ = None

    def fit(self, x_vector: DataFrame, y_vector=None) -> PFISelector:
        x_train, x_test, y_train, y_test = train_test_split(
            x_vector, y_vector, test_size=0.1, random_state=42
        )
        self.estimator_ = self.estimator.fit(x_train, y_train)
        results = permutation_importance(
            self.estimator_,
            x_test,
            y_test,
            n_repeats=self.n_repeats,
            random_state=42,
            n_jobs=-1,
        )

        self.feature_importances_ = results.importances_mean

        threshold_value = percentile(self.feature_importances_, 75)

        self.important_indices_ = self.feature_importances_ > threshold_value
        self.important_features_ = x_vector.columns[self.important_indices_]
        return self

    def transform(self, x_vector: DataFrame) -> DataFrame:
        return x_vector[self.important_features_]

    def get_support(self, *_) -> list[bool]:
        return self.important_indices_.to_list()



from __future__ import annotations

import asyncio
from typing import Callable

from aiomqtt import MqttError

from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.events.events import BaseEvent


class EventBus:
    def __init__(self) -> None:
        self._listeners: dict[str, list[Callable]] = {}
        self._subscribed_events: dict[str, type[BaseEvent]] = {}
        self._tasks: set[asyncio.Task] = set()

    def subscribe(
        self,
        event: type[BaseEvent] | list[type[BaseEvent]],
        listener: Callable,
    ) -> None:
        if isinstance(event, list):
            for _event in event:
                self.subscribe(_event, listener)
            return

        event_key = event.event_key()

        logger.info(f"Event subscribed: {event_key}")

        if event_key not in self._listeners:
            self._listeners[event_key] = []
        self._listeners[event_key].append(listener)
        self._subscribed_events[event_key] = event

    @property
    def subcribed_events(self) -> list[type[BaseEvent]]:
        return list(self._subscribed_events.values())

    def unsubscribe(self, event: type[BaseEvent], listener: Callable) -> None:
        event_key = event.event_key()
        if event_key in self._listeners:
            self._listeners[event_key].remove(listener)
            if not self._listeners[event_key]:
                self._listeners.pop(event_key, None)
                self._subscribed_events.pop(event_key, None)

    async def emit(self, event: BaseEvent) -> None:
        try:
            event_key = event.event_key()
            logger.trace(f"Event emitted: {event_key}")

            if event_key in self._listeners:
                if event.AWAIT:
                    await self._notify_listeners(event, self._listeners[event_key])
                else:
                    task = asyncio.create_task(
                        self._notify_listeners(event, self._listeners[event_key])
                    )
                    self._tasks.add(task)
                    task.add_done_callback(self._tasks.remove)
        except asyncio.CancelledError:
            pass

    async def _notify_listeners(
        self, event: BaseEvent, listeners: list[Callable]
    ) -> None:
        try:
            await asyncio.gather(
                *[self._notify_listener(listener, event) for listener in listeners]
            )
        except MqttError as error:
            raise error
        except asyncio.CancelledError:
            pass

    async def _notify_listener(self, listener: Callable, event: BaseEvent) -> None:
        try:
            await listener(event)
        except InvalidDataException as error:
            logger.warning("{message}, skipping this loop", message=error.message)

    def cancel_tasks(self):
        for task in self._tasks:
            try:
                task.cancel()
            finally:
                pass



import json

from pydantic import BaseModel

from solaredge2mqtt.core.events.events import BaseEvent


class MQTTPublishEvent(BaseEvent):
    AWAIT = True

    def __init__(
        self,
        topic: str,
        payload: str | int | float | BaseModel,
        retain: bool = False,
        qos: int = 0,
        topic_prefix: str | None = None,
        exclude_none: bool = False,
    ):
        self._topic: str = topic
        self._payload: str | int | float | BaseModel = payload
        self._retain: bool = retain
        self._qos: int = qos
        self._topic_prefix: str | None = topic_prefix
        self._exclude_none: bool = exclude_none

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def payload(self) -> str | int | float | BaseModel:
        return self._payload

    @property
    def retain(self) -> bool:
        return self._retain

    @property
    def qos(self) -> int:
        return self._qos

    @property
    def topic_prefix(self) -> str | None:
        return self._topic_prefix

    @property
    def exclude_none(self) -> bool:
        return self._exclude_none


class MQTTReceivedEvent(BaseEvent):
    def __init__(self, topic: str, payload: str):
        self._topic: str = topic
        self._payload: str = payload

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def payload(self) -> str:
        return self._payload

    @property
    def json(self) -> dict | str | int | float:
        return json.loads(self.payload)


class MQTTSubscribeEvent(BaseEvent):
    AWAIT = True

    def __init__(self, topic: str):
        self._topic: str = topic

    @property
    def topic(self) -> str:
        return self._topic



"""
    This module, service.py, is part of the SolarEdge2MQTT service, which reads data 
    from a SolarEdge inverter and publishes it to an MQTT broker. It uses the asyncio 
    library for asynchronous I/O and the aiomqtt library for MQTT communication. 
    The module also includes a run function to initialize and start the service.
"""

import asyncio as aio
import platform
import signal
from time import time
from typing import Callable

from aiomqtt import MqttError
from tzlocal import get_localzone_name

from solaredge2mqtt import __version__
from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import ConfigurationException
from solaredge2mqtt.core.influxdb import InfluxDB
from solaredge2mqtt.core.logging import initialize_logging, logger
from solaredge2mqtt.core.mqtt import MQTTClient
from solaredge2mqtt.core.settings import service_settings
from solaredge2mqtt.core.timer import Timer
from solaredge2mqtt.services.energy import EnergyService
from solaredge2mqtt.services.forecast import ForecastService
from solaredge2mqtt.services.homeassistant import HomeAssistantDiscovery
from solaredge2mqtt.services.monitoring import MonitoringSite
from solaredge2mqtt.services.powerflow import PowerflowService
from solaredge2mqtt.services.weather import WeatherClient

LOCAL_TZ = get_localzone_name()


def run():
    try:
        service = Service()
        loop = aio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, service.cancel)
        loop.add_signal_handler(signal.SIGTERM, service.cancel)
        loop.run_until_complete(service.main_loop())
    except ConfigurationException:
        logger.error("Configuration error")
    except aio.exceptions.CancelledError:
        logger.debug("Service cancelled")
    finally:
        loop.close()


class Service:
    def __init__(self):
        self.settings = service_settings()
        initialize_logging(self.settings.logging_level)
        logger.debug(self.settings)

        self.event_bus = EventBus()
        self.timer = Timer(self.event_bus, self.settings.interval)

        self.mqtt = MQTTClient(self.settings.mqtt, self.event_bus)

        self.cancel_request = aio.Event()
        self.loops: set[aio.Task] = set()

        self.influxdb: InfluxDB | None = (
            InfluxDB(self.settings.influxdb, self.settings.prices, self.event_bus)
            if self.settings.is_influxdb_configured
            else None
        )

        self.energy: EnergyService | None = (
            EnergyService(self.event_bus, self.influxdb)
            if self.settings.is_influxdb_configured
            else None
        )

        self.powerflow = PowerflowService(self.settings, self.event_bus, self.influxdb)

        self.monitoring: MonitoringSite | None = (
            MonitoringSite(self.settings.monitoring, self.event_bus)
            if self.settings.is_monitoring_configured
            else None
        )

        self.weather: WeatherClient | None = (
            WeatherClient(self.settings, self.event_bus)
            if self.settings.is_weather_configured
            else None
        )

        self.forecast: ForecastService | None = (
            ForecastService(
                self.settings.forecast,
                self.settings.location,
                self.event_bus,
                self.influxdb,
            )
            if self.settings.is_forecast_configured
            else None
        )

        self.homeassistant: HomeAssistantDiscovery | None = (
            HomeAssistantDiscovery(self.settings, self.event_bus)
            if self.settings.is_homeassistant_configured
            else None
        )

    def cancel(self):
        logger.info("Stopping SolarEdge2MQTT service...")
        self.cancel_request.set()
        for tasks in aio.all_tasks():
            tasks.cancel()

    async def main_loop(self):
        logger.info("Starting SolarEdge2MQTT service...")
        logger.info("Version: {version}", version=__version__)
        logger.info(
            f"Operationg system: {platform.platform()} ({platform.system()}/{platform.machine()})"
        )
        logger.debug(self.settings)
        logger.info("Timezone: {timezone}", timezone=LOCAL_TZ)

        if self.settings.is_influxdb_configured:
            self.influxdb.initialize_buckets()

        while not self.cancel_request.is_set():
            try:
                async with self.mqtt:
                    await self.mqtt.publish_status_online()

                    if self.settings.is_homeassistant_configured:
                        await self.homeassistant.async_init()

                    self._start_mqtt_listener()
                    self.schedule_loop(1, self.timer.loop)

                    await aio.gather(*self.loops)
            except MqttError:
                logger.error("MQTT error, reconnecting in 5 seconds...")
                await aio.sleep(5)
            except aio.exceptions.CancelledError:
                logger.debug("Loops cancelled")
                return
            finally:
                await self.finalize()

    async def finalize(self):
        try:
            await self.mqtt.publish_status_offline()
        except MqttError:
            pass

        try:
            self.event_bus.cancel_tasks()
        finally:
            pass

        for task in self.loops:
            try:
                task.cancel()
            finally:
                pass

    def _start_mqtt_listener(self):
        task = aio.create_task(self.mqtt.listen())
        self.loops.add(task)
        task.add_done_callback(self.loops.remove)

    def schedule_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: int = 0,
        args: list[any] = None,
    ):
        loop = aio.create_task(
            self.run_loop(interval_in_seconds, handles, delay_start, args)
        )
        self.loops.add(loop)
        loop.add_done_callback(self.loops.remove)

    async def run_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: int = 0,
        args: list[any] = None,
    ):
        if not isinstance(handles, list):
            handles = [handles]

        await aio.sleep(delay_start)

        while not self.cancel_request.is_set():
            execution_time = 0
            start_time = time()
            for handle in handles:
                await handle(*args or [])
            execution_time = time() - start_time

            if execution_time < interval_in_seconds:
                await aio.sleep(interval_in_seconds - execution_time)
            else:
                await aio.sleep(interval_in_seconds)



import base64
import hashlib

from pydantic import BaseModel, Field, computed_field

from solaredge2mqtt.core.models import EnumModel


class HomeAssistantBaseModel(BaseModel):
    client_id: str = Field(..., exclude=True)

    def hash_unique_id(self, ids: list[str | int]) -> str:
        ids.append(self.client_id)
        unique_id = "_".join([str(id) for id in ids])

        hash_obj = hashlib.sha256(unique_id.encode())
        hash_digest = hash_obj.digest()

        base64_encoded = base64.urlsafe_b64encode(hash_digest).decode()

        return base64_encoded[:10]


class HomeAssistantDevice(HomeAssistantBaseModel):
    name: str
    state_topic: str = Field(exclude=True)
    manufacturer: str | None = Field(None)
    model: str | None = Field(None)
    hw_version: str | None = Field(None)
    serial_number: str | None = Field(None)
    sw_version: str | None = Field(None)
    via_device: str | None = Field(None)

    @computed_field
    @property
    def identifiers(self) -> str:
        return self.hash_unique_id(
            [self.name, self.manufacturer, self.model, self.serial_number]
        )


class HomeAssistantEntityType(EnumModel):
    APPARENT_POWER = "apparent_power", "sensor", "apparent_power", "measurement", "VA"
    BATTERY = "battery", "sensor", "battery", "measurement", "%"
    CURRENT_A = "current_a", "sensor", "current", "measurement", "A"
    ENERGY_KWH = "energy_kwh", "sensor", "energy", "total_increasing", "kWh"
    ENERGY_WH = "energy_wh", "sensor", "energy", "total_increasing", "Wh"
    FREQUENCY_HZ = "frequency_hz", "sensor", "frequency", "measurement", "Hz"
    MONETARY = "monetary", "sensor", "monetary", None, None
    PERCENTAGE = "percentage", "sensor", None, "measurement", "%"
    PLUG = "plug", "binary_sensor", "plug", None, None
    POWER_FACTOR = "power_factor", "sensor", "power_factor", "measurement", "%"
    POWER_KW = "power_kw", "sensor", "power", "measurement", "kW"
    POWER_W = "power_w", "sensor", "power", "measurement", "W"
    REACTIVE_POWER = "reactive_power", "sensor", "reactive_power", "measurement", "VAr"
    VOLTAGE_V = "voltage_v", "sensor", "voltage", "measurement", "V"

    def __init__(
        self,
        key: str,
        typed: str,
        device_class: str | None = None,
        state_class: str | None = None,
        unit_of_measurement: str | None = None,
    ):
        self._key: str = key
        self._typed: str = typed
        self._device_class: str = device_class
        self._state_class: str | None = state_class
        self._unit_of_measurement: str | None = unit_of_measurement

    @property
    def typed(self) -> str:
        return self._typed

    @property
    def device_class(self) -> str:
        return self._device_class

    @property
    def state_class(self) -> str | None:
        return self._state_class

    @property
    def unit_of_measurement(self) -> str | None:
        return self._unit_of_measurement

    def field(
        self, title: str | None = None, icon: str | None = None
    ) -> dict[str, any]:
        json_schema_extra = {
            "ha_type": self,
            "icon": icon,
        }

        return {"title": title, "json_schema_extra": json_schema_extra}


class HomeAssistantEntity(HomeAssistantBaseModel):
    name: str
    device: HomeAssistantDevice
    _icon: str | None = None
    jsonpath: str | None = Field(None, exclude=True)
    ha_type: HomeAssistantEntityType = Field(exclude=True)
    unit: str | None = Field(None, exclude=True)

    def __init__(self, device: HomeAssistantDevice, icon: str | None = None, **data):
        super().__init__(client_id=device.client_id, device=device, **data)

        self._icon = icon

    @computed_field
    @property
    def unique_id(self) -> str:
        return self.hash_unique_id(
            [self.device.identifiers, self.name, self.state_topic, self.value_template]
        )

    @computed_field
    @property
    def state_topic(self) -> str:
        return self.device.state_topic

    @computed_field
    @property
    def value_template(self) -> str | None:
        return f"{{{{ value_json.{self.jsonpath} }}}}" if self.jsonpath else None

    @computed_field
    @property
    def state_class(self) -> str | None:
        return self.ha_type.state_class

    @computed_field
    @property
    def device_class(self) -> str | None:
        return self.ha_type.device_class

    @computed_field
    @property
    def unit_of_measurement(self) -> str | None:
        return self.ha_type.unit_of_measurement if self.unit is None else self.unit

    @computed_field
    @property
    def payload_on(self) -> str | None:
        return "true" if self.ha_type.typed == "binary_sensor" else None

    @computed_field
    @property
    def payload_off(self) -> str | None:
        return "false" if self.ha_type.typed == "binary_sensor" else None

    @computed_field
    @property
    def icon(self) -> str | None:
        return f"mdi:{self._icon}" if self._icon else None



from __future__ import annotations

from typing import TYPE_CHECKING

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.influxdb import InfluxDB
from solaredge2mqtt.core.influxdb.events import InfluxDBAggregatedEvent
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.services.energy.events import EnergyReadEvent
from solaredge2mqtt.services.energy.models import (
    HistoricEnergy,
    HistoricPeriod,
    HistoricQuery,
)


class EnergyService:
    def __init__(
        self,
        event_bus: EventBus,
        influxdb: InfluxDB,
    ):
        self.influxdb = influxdb

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(InfluxDBAggregatedEvent, self.read_historic_energy)

    async def read_historic_energy(self, _) -> None:
        for period in HistoricPeriod:
            record = self.influxdb.query_timeunit(period, "energy")
            if record is None:
                if period.query == HistoricQuery.LAST:
                    logger.info(
                        "No data found for {period}, skipping this loop", period=period
                    )
                else:
                    raise InvalidDataException(f"No energy data for {period}")

                continue

            energy = HistoricEnergy(record, period)

            logger.info(
                "Read from influxdb {period} energy: {energy.pv_production} kWh",
                period=period,
                energy=energy,
            )

            await self.event_bus.emit(EnergyReadEvent(energy))
            await self.event_bus.emit(MQTTPublishEvent(energy.mqtt_topic(), energy))



import json

from pymodbus.exceptions import ModbusException
from solaredge_modbus import Inverter

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.services.modbus.events import (
    ModbusBatteriesReadEvent,
    ModbusInverterReadEvent,
    ModbusMetersReadEvent,
)
from solaredge2mqtt.services.modbus.models import (
    SunSpecBattery,
    SunSpecInverter,
    SunSpecMeter,
)
from solaredge2mqtt.services.modbus.settings import ModbusSettings

SunSpecRawData = dict[str, str | int]

LOGGING_DEVICE_INFO = "{device} ({info.manufacturer} {info.model} {info.serialnumber})"


class Modbus:
    inverter: Inverter

    def __init__(self, settings: ModbusSettings, event_bus: EventBus):
        self.inverter = Inverter(
            host=settings.host,
            port=settings.port,
            timeout=settings.timeout,
            unit=settings.unit,
        )

        logger.info(
            "Using SolarEdge inverter via modbus: {host}:{port}",
            host=settings.host,
            port=settings.port,
        )

        self.event_bus = event_bus

    async def get_data(
        self,
    ) -> (
        tuple[
            SunSpecInverter | None,
            dict[str, SunSpecMeter] | None,
            dict[str, SunSpecBattery | None],
        ]
        | None
    ):
        inverter_data = None
        meters_data = None
        batteries_data = None

        try:
            inverter_raw, meters_raw, batteries_raw = self._get_raw_data()

            inverter_data = self._map_inverter(inverter_raw)
            meters_data = self._map_meters(meters_raw)
            batteries_data = self._map_batteries(batteries_raw)
        except (ModbusException, KeyError) as error:
            raise InvalidDataException("Invalid modbus data") from error

        await self.event_bus.emit(ModbusInverterReadEvent(inverter_data))
        if meters_data:
            await self.event_bus.emit(ModbusMetersReadEvent(meters_data))
        if batteries_data:
            await self.event_bus.emit(ModbusBatteriesReadEvent(batteries_data))

        return inverter_data, meters_data, batteries_data

    def _get_raw_data(
        self,
    ) -> tuple[SunSpecRawData, dict[str, SunSpecRawData], dict[str, SunSpecRawData]]:
        inverter_raw = self.inverter.read_all()
        meters_raw = {
            meter_key: meter_obj.read_all()
            for meter_key, meter_obj in self.inverter.meters().items()
        }
        batteries_raw = {
            battery_key: battery_raw
            for battery_key, battery_obj in self.inverter.batteries().items()
            if (battery_raw := battery_obj.read_all())
        }

        return inverter_raw, meters_raw, batteries_raw

    def _map_inverter(self, inverter_raw: SunSpecRawData) -> SunSpecInverter:
        logger.debug(
            "Inverter raw:\n{raw}",
            raw=json.dumps(inverter_raw, indent=4),
        )

        inverter_data = SunSpecInverter(inverter_raw)
        logger.debug(inverter_data)
        logger.info(
            LOGGING_DEVICE_INFO
            + ": {status}, AC {power_ac} W, DC {power_dc} W, {energytotal} kWh",
            device="Inverter",
            info=inverter_data.info,
            status=inverter_data.status,
            power_ac=inverter_data.ac.power.actual,
            power_dc=inverter_data.dc.power,
            energytotal=round(inverter_data.energytotal / 1000, 2),
        )

        return inverter_data

    def _map_meters(
        self, meters_raw: dict[str, SunSpecRawData]
    ) -> dict[str, SunSpecInverter]:
        meters = {}
        for meter_key, meter_raw in meters_raw.items():
            logger.debug(
                "Meter {meter} raw:\n{raw}",
                meter=meter_key,
                raw=json.dumps(meter_raw, indent=4),
            )

            meter_data = SunSpecMeter(meter_raw)
            logger.debug(meter_data)
            logger.info(
                LOGGING_DEVICE_INFO + ": {power} W, {consumption} kWh, {delivery} kWh",
                device=meter_key,
                info=meter_data.info,
                power=meter_data.power.actual,
                consumption=round(meter_data.energy.totalimport / 1000, 3),
                delivery=round(meter_data.energy.totalexport / 1000, 3),
            )

            meters[meter_key] = meter_data

        return meters

    def _map_batteries(
        self, batteries_raw: dict[str, SunSpecRawData]
    ) -> dict[str, SunSpecInverter]:
        batteries = {}
        for battery_key, battery_raw in batteries_raw.items():
            logger.debug(
                "Battery {battery} raw:\n{raw}",
                battery=battery_key,
                raw=json.dumps(battery_raw, indent=4),
            )

            battery_data = SunSpecBattery(battery_raw)
            logger.debug(battery_data)
            logger.info(
                LOGGING_DEVICE_INFO + ": {power} W, {state_of_charge} %",
                device=battery_key,
                info=battery_data.info,
                power=battery_data.power,
                state_of_charge=battery_data.state_of_charge,
            )

            batteries[battery_key] = battery_data

        return batteries



from aiomqtt import Client, Will
from pydantic import BaseModel

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import (
    MQTTPublishEvent,
    MQTTReceivedEvent,
    MQTTSubscribeEvent,
)
from solaredge2mqtt.core.mqtt.settings import MQTTSettings


class MQTTClient(Client):
    def __init__(self, settings: MQTTSettings, event_bus: EventBus):
        self.broker = settings.broker
        self.port = settings.port

        self.topic_prefix = settings.topic_prefix

        logger.info(
            "Using MQTT broker: {broker}:{port}",
            broker=settings.broker,
            port=settings.port,
        )

        will = Will(
            topic=f"{self.topic_prefix}/status", payload="offline", qos=1, retain=True
        )

        self._subscribed_topics = set()

        self.event_bus = event_bus
        self._subscribe_events()

        super().__init__(
            self.broker,
            self.port,
            username=settings.username,
            password=settings.password.get_secret_value(),
            identifier=settings.client_id,
            will=will,
        )

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(MQTTPublishEvent, self.event_listener)
        self.event_bus.subscribe(MQTTSubscribeEvent, self._subscribe_topic)

    async def _subscribe_topic(self, event: MQTTSubscribeEvent) -> None:
        if event.topic not in self._subscribed_topics:
            logger.info(f"Subscribing to topic: {event.topic}")
            self._subscribed_topics.add(event.topic)
            await self.subscribe(event.topic)

    async def listen(self) -> None:
        if self._subscribed_topics:
            logger.info("MQTT listen on subscribed topics")
            async for message in self.messages:
                logger.debug(
                    "MQTT message topic: {message.topic}={message.payload}",
                    message=message,
                )
                payload = message.payload.decode()
                await self.event_bus.emit(
                    MQTTReceivedEvent(str(message.topic), payload)
                )

    async def publish_status_online(self) -> None:
        await self.publish_to("status", "online", True)

    async def publish_status_offline(self) -> None:
        await self.publish_to("status", "offline", True)

    async def event_listener(self, event: MQTTPublishEvent) -> None:
        await self.publish_to(
            event.topic,
            event.payload,
            event.retain,
            event.qos,
            event.topic_prefix,
            event.exclude_none,
        )

    async def publish_to(
        self,
        topic: str,
        payload: str | int | float | BaseModel,
        retain: bool = False,
        qos: int = 1,
        topic_prefix: str | None = None,
        exclude_none: bool = False,
    ) -> None:
        if self._connected:
            topic = f"{topic_prefix or self.topic_prefix}/{topic}"

            if isinstance(payload, BaseModel):
                payload = payload.model_dump_json(exclude_none=exclude_none)

            await self.publish(topic, payload, qos=qos, retain=retain)



from pydantic import BaseModel, Field, SecretStr


class MQTTSettings(BaseModel):
    client_id: str = Field("solaredge2mqtt")
    broker: str
    port: int = Field(1883)
    username: str
    password: SecretStr
    topic_prefix: str = Field("solaredge")



from __future__ import annotations

from typing import TYPE_CHECKING

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import (
    MQTTPublishEvent,
    MQTTReceivedEvent,
    MQTTSubscribeEvent,
)
from solaredge2mqtt.services.energy.events import EnergyReadEvent
from solaredge2mqtt.services.events import ComponentEvent, ComponentsEvent
from solaredge2mqtt.services.forecast.events import ForecastEvent
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantDevice,
    HomeAssistantEntity,
    HomeAssistantEntityType,
)
from solaredge2mqtt.services.modbus.events import (
    ModbusBatteriesReadEvent,
    ModbusInverterReadEvent,
    ModbusMetersReadEvent,
)
from solaredge2mqtt.services.models import Component
from solaredge2mqtt.services.powerflow.events import PowerflowGeneratedEvent
from solaredge2mqtt.services.wallbox.events import WallboxReadEvent

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings import ServiceSettings


class HomeAssistantDiscovery:
    def __init__(self, service_settings: ServiceSettings, event_bus: EventBus) -> None:
        self.settings = service_settings
        self._send_entities: dict[str, HomeAssistantEntity] = {}
        logger.info("Home Assistant discovery enabled")

        self._status_topic = f"{self.settings.homeassistant.topic_prefix}/status"

        self._seen_energy_periods: set[str] = set()

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(
            [
                ForecastEvent,
                EnergyReadEvent,
                ModbusInverterReadEvent,
                PowerflowGeneratedEvent,
                WallboxReadEvent,
            ],
            self.component_discovery,
        )

        self.event_bus.subscribe(
            [ModbusMetersReadEvent, ModbusBatteriesReadEvent], self.components_discovery
        )

        self.event_bus.subscribe(
            MQTTReceivedEvent,
            self.homeassistant_status,
        )

    async def async_init(self) -> None:
        await self.event_bus.emit(MQTTSubscribeEvent(self._status_topic))

    async def component_discovery(self, event: ComponentEvent) -> None:
        publish = True
        if isinstance(event, EnergyReadEvent):
            period = event.component.info.period
            publish = (
                period.auto_discovery and period.topic not in self._seen_energy_periods
            )
            self._seen_energy_periods.add(period.topic)
        else:
            self.event_bus.unsubscribe(event, self.component_discovery)

        if publish:
            logger.info(f"Home Assistant discovery component: {event.component}")
            await self.publish_component(event.component)

    async def components_discovery(self, event: ComponentsEvent) -> None:
        self.event_bus.unsubscribe(event, self.components_discovery)
        for name, component in event.components.items():
            logger.info(f"Home Assistant discovery component: {component}")
            await self.publish_component(component, name)

    async def publish_component(self, component: Component, name: str = "") -> None:
        subtopic = f"/{name.lower()}" if name else ""

        state_topic = (
            f"{self.settings.mqtt.topic_prefix}/{component.mqtt_topic()}{subtopic}"
        )

        if name:
            device_info = component.homeassistant_device_info_with_name(name)
        else:
            device_info = component.homeassistant_device_info()

        logger.trace(device_info)
        device = HomeAssistantDevice(
            client_id=self.settings.mqtt.client_id,
            state_topic=state_topic,
            **device_info,
        )
        logger.debug(device)

        entities_info = component.homeassistant_entities_info()
        for entity_info in entities_info:
            if self.settings.is_prices_configured and entity_info["ha_type"] == str(
                HomeAssistantEntityType.MONETARY
            ):
                entity_info["unit"] = self.settings.prices.currency

            entity = HomeAssistantEntity(
                device=device,
                **entity_info,
            )
            topic = (
                f"{entity.ha_type.typed}/"
                f"{state_topic.replace('/', '_')}/"
                f"{entity.unique_id}/config"
            )
            logger.debug(entity)

            self._send_entities[topic] = entity

            await self.event_bus.emit(
                MQTTPublishEvent(
                    topic=topic,
                    payload=entity,
                    topic_prefix=self.settings.homeassistant.topic_prefix,
                    exclude_none=True,
                )
            )

    async def homeassistant_status(self, event: MQTTReceivedEvent) -> None:
        if event.topic == self._status_topic and event.payload == "online":
            logger.info("Home Assistant status changed to online resend discovery")
            for topic, entity in self._send_entities.items():
                await self.event_bus.emit(
                    MQTTPublishEvent(
                        topic=topic,
                        payload=entity,
                        topic_prefix=self.settings.homeassistant.topic_prefix,
                        exclude_none=True,
                    )
                )



from solaredge2mqtt.services.events import ComponentEvent, ComponentsEvent


class ModbusInverterReadEvent(ComponentEvent):
    pass


class ModbusMetersReadEvent(ComponentsEvent):
    pass


class ModbusBatteriesReadEvent(ComponentsEvent):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from requests.exceptions import HTTPError

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.services.http import HTTPClient
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import Interval10MinTriggerEvent
from solaredge2mqtt.services.weather.events import WeatherUpdateEvent
from solaredge2mqtt.services.weather.models import OpenWeatherMapOneCall

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings import ServiceSettings

ONECALL_URL = "https://api.openweathermap.org/data/3.0/onecall"
TIMEMACHINE_URL = "https://api.openweathermap.org/data/3.0/onecall/timemachine"


class WeatherClient(HTTPClient):
    def __init__(self, settings: ServiceSettings, event_bus: EventBus) -> None:
        super().__init__("Weather API")

        self.location = settings.location
        self.settings = settings.weather

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self):
        self.event_bus.subscribe(Interval10MinTriggerEvent, self.loop)

    async def loop(self, _):
        weather = self.get_weather()
        await self.event_bus.emit(WeatherUpdateEvent(weather))
        await self.event_bus.emit(MQTTPublishEvent("weather/current", weather.current))

    def get_weather(self) -> OpenWeatherMapOneCall:
        try:
            logger.info("Reading weather data from OpenWeatherMap")
            result = self._get(
                ONECALL_URL,
                params={
                    "lat": self.location.latitude,
                    "lon": self.location.longitude,
                    "exclude": "minutely,daily,alerts",
                    "units": "metric",
                    "lang": self.settings.language,
                    "appid": self.settings.api_key.get_secret_value(),
                },
                timeout=7,
            )

            logger.trace(result)

            if result is None:
                raise InvalidDataException(
                    "Unable to read weather data from OpenWeatherMap"
                )

            weather = OpenWeatherMapOneCall(**result)

            return weather
        except HTTPError as error:
            status_code = error.response.status_code
            if status_code == 401:
                error_msg = (
                    "Invalid OpenWeatherMap API key or no subscription to OneCall-API"
                )
            else:
                error_msg = "Unable to read weather data from OpenWeatherMap (HTTP {status_code})"

            raise InvalidDataException(error_msg) from error



from __future__ import annotations
from typing import TYPE_CHECKING

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.core.influxdb import InfluxDB
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import IntervalBaseTriggerEvent
from solaredge2mqtt.services.modbus import Modbus
from solaredge2mqtt.services.modbus.models import SunSpecBattery
from solaredge2mqtt.services.powerflow.events import PowerflowGeneratedEvent
from solaredge2mqtt.services.powerflow.models import Powerflow
from solaredge2mqtt.services.wallbox import WallboxClient

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings.models import ServiceSettings


class PowerflowService:

    def __init__(
        self,
        settings: ServiceSettings,
        event_bus: EventBus,
        influxdb: InfluxDB | None = None,
    ):
        self.settings = settings

        self.influxdb = influxdb

        self.modbus = Modbus(self.settings.modbus, event_bus)

        self.wallbox = (
            WallboxClient(self.settings.wallbox, event_bus)
            if self.settings.is_wallbox_configured
            else None
        )

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(IntervalBaseTriggerEvent, self.calculate_powerflow)

    async def calculate_powerflow(self, _) -> None:
        inverter_data, meters_data, batteries_data = await self.modbus.get_data()

        if any(data is None for data in [inverter_data, meters_data, batteries_data]):
            raise InvalidDataException("Invalid modbus data")

        for battery in batteries_data.values():
            if not battery.is_valid:
                logger.debug(battery)
                raise InvalidDataException("Invalid battery data")

        evcharger = 0
        wallbox_data = None
        if self.settings.is_wallbox_configured:
            try:
                wallbox_data = await self.wallbox.get_data()
                logger.trace(
                    "Wallbox: {wallbox_data.power} W", wallbox_data=wallbox_data
                )
                evcharger = wallbox_data.power
            except ConfigurationException as ex:
                logger.warning(f"{ex.component}: {ex.message}")

        powerflow = Powerflow.from_modbus(
            inverter_data, meters_data, batteries_data, evcharger
        )
        if not powerflow.is_valid:
            logger.info(powerflow)
            raise InvalidDataException("Invalid powerflow data")

        if Powerflow.is_not_valid_with_last(powerflow):
            logger.debug(powerflow)
            raise InvalidDataException("Value change not valid, skipping this loop")

        self.write_to_influxdb(batteries_data, powerflow)

        logger.debug(powerflow)
        logger.info(
            "Powerflow: PV {pv_production} W, Inverter {inverter.power} W, "
            + "House {consumer.house} W, Grid {grid.power} W, "
            + "Battery {battery.power} W, Wallbox {consumer.evcharger} W",
            pv_production=powerflow.pv_production,
            inverter=powerflow.inverter,
            consumer=powerflow.consumer,
            grid=powerflow.grid,
            battery=powerflow.battery,
        )

        await self.event_bus.emit(
            MQTTPublishEvent(inverter_data.mqtt_topic(), inverter_data)
        )

        for key, component in {**meters_data, **batteries_data}.items():
            await self.event_bus.emit(
                MQTTPublishEvent(
                    f"{component.mqtt_topic()}/{key.lower()}",
                    component,
                )
            )

        if wallbox_data is not None:
            await self.event_bus.emit(
                MQTTPublishEvent(wallbox_data.mqtt_topic(), wallbox_data)
            )

        await self.event_bus.emit(MQTTPublishEvent(powerflow.mqtt_topic(), powerflow))

        await self.event_bus.emit(PowerflowGeneratedEvent(powerflow))

    def write_to_influxdb(
        self, batteries_data: dict[str, SunSpecBattery], powerflow: Powerflow
    ):
        if self.influxdb is not None:
            points = [powerflow.prepare_point()]

            for battery in batteries_data.values():
                points.append(battery.prepare_point())

            self.influxdb.write_points(points)



from requests.exceptions import HTTPError

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import Interval10MinTriggerEvent
from solaredge2mqtt.services.http import HTTPClient
from solaredge2mqtt.services.monitoring.models import (
    LogicalInfo,
    LogicalInverter,
    LogicalModule,
    LogicalString,
)
from solaredge2mqtt.services.monitoring.settings import MonitoringSettings

LOGIN_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/login"
LOGICAL_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/sites/{site_id}/layout/logical"


class MonitoringSite(HTTPClient):
    def __init__(self, settings: MonitoringSettings, event_bus: EventBus) -> None:
        super().__init__("Monitoring Site")
        self.settings = settings

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(Interval10MinTriggerEvent, self.get_data)

    async def get_data(self, _):
        modules = self.get_module_energies()

        if modules is None:
            logger.warning("Invalid monitoring data, skipping this loop")
            return

        energy_total = 0
        count_modules = 0
        for module in modules:
            if module.energy is not None:
                count_modules += 1
                energy_total += module.energy

        logger.info(
            "Read from monitoring total energy: {energy_total} kWh from {count_modules} modules",
            energy_total=energy_total / 1000,
            count_modules=count_modules,
        )

        await self.event_bus.emit(
            MQTTPublishEvent("monitoring/pv_energy_today", energy_total)
        )

        for module in modules:
            await self.event_bus.emit(
                MQTTPublishEvent(
                    f"monitoring/module/{module.info.serialnumber}",
                    module,
                )
            )

    def login(self) -> None:
        try:
            self._post(
                LOGIN_URL,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "j_username": self.settings.username,
                    "j_password": self.settings.password.get_secret_value(),
                },
                timeout=10,
                expect_json=False,
            )

            logger.info("Login to monitoring site successful")
        except HTTPError as error:
            raise ConfigurationException(
                "Monitoring", "Unable to login to monitoring account"
            ) from error

    def get_module_energies(self) -> list[LogicalInverter] | None:
        modules = None
        logical = self._get_logical()

        if logical is None:
            raise InvalidDataException(
                "Unable to read logical layout from monitoring site"
            )

        inverters = self._parse_inverters(
            logical["logicalTree"]["children"], logical["reportersData"]
        )

        modules = []

        for inverter in inverters:
            logger.debug(
                "Inverter: {inverter}", inverter=inverter.model_dump_json(indent=4)
            )
            for string in inverter.strings:
                for module in string.modules:
                    modules.append(module)

        return modules

    def _get_logical(self) -> dict | None:
        result = None
        try:
            if "CSRF-TOKEN" not in self.session.cookies:
                self.login()

            result = self._get(
                LOGICAL_URL.format(site_id=self.settings.site_id),
                headers={
                    "Content-Type": "application/x-www-form-urlencoded",
                    "X-CSRF-TOKEN": self.session.cookies["CSRF-TOKEN"],
                },
                timeout=10,
            )
        except HTTPError as error:
            raise InvalidDataException("Unable to read logical layout") from error

        return result

    def _parse_inverters(self, inverter_objs, reporters_data) -> list[LogicalInverter]:
        inverters = []

        for inverter_obj in inverter_objs:
            info = LogicalInfo.map(inverter_obj["data"])
            if "INVERTER" in info["type"]:
                inverter = LogicalInverter(
                    info=info,
                    energy=(
                        reporters_data[info["id"]]["unscaledEnergy"]
                        if info["id"] in reporters_data
                        else None
                    ),
                )

                self._parse_strings(inverter, inverter_obj["children"], reporters_data)

                inverters.append(inverter)

            else:
                logger.info("Unknown inverter type: {type}", type=info["type"])

        return inverters

    def _parse_strings(self, inverter, string_objs, reporters_data):
        for string_obj in string_objs:
            info = LogicalInfo.map(string_obj["data"])
            string = LogicalString(
                info=info,
                energy=(
                    reporters_data[info["id"]]["unscaledEnergy"]
                    if info["id"] in reporters_data
                    else None
                ),
            )

            self._parse_panels(string, string_obj["children"], reporters_data)

            inverter.strings.append(string)

    def _parse_panels(self, string, panel_objs, reporters_data):
        for panel_obj in panel_objs:
            info = LogicalInfo.map(panel_obj["data"])
            panel = LogicalModule(
                info=info,
                energy=(
                    reporters_data[info["id"]]["unscaledEnergy"]
                    if info["id"] in reporters_data
                    else None
                ),
            )

            string.modules.append(panel)
